local
  open Fnlib Mixture Const Smlprim Globals Location Units;
in

datatype reason = 
    UnifyCircular | UnifyEquality | UnifyExplicit
  | UnifyTup | UnifyRec of Lab | UnifyOther

exception Unify of reason;

val tyname_unit     : TyName;
val tyname_bool     : TyName;
val tyname_int      : TyName;
val tyname_word     : TyName;
val tyname_word8    : TyName;
val tyname_char     : TyName;
val tyname_real     : TyName;
val tyname_string   : TyName;
val tyname_substring: TyName;
val tyname_syserror : TyName;
val tyname_list     : TyName;
val tyname_vector   : TyName;
val tyname_ref      : TyName;
val tyname_exn      : TyName;
val tyname_option   : TyName;
val tyname_order : TyName;
val tyname_frag     : TyName;
val tyname_ppstream : TyName;

val mkConInfo : unit -> ConInfo;
val mkExConInfo : unit -> ExConInfo;
val mkPrimInfo : int -> SMLPrim -> PrimInfo;
val isExConStatic : ExConInfo -> bool;

val normType: Type -> Type;

val resetTypePrinter: unit -> unit;
val collectExplicitVars: Type -> unit;
val printNextType: Type -> unit;
val printType: Type -> unit;
val printScheme: TypeScheme -> unit;

val newExplicitTypeVar: string -> TypeVar;
val mkTypeVar   : bool -> bool -> bool -> int -> TypeVar;
val newTypeVar  : bool -> bool -> bool -> TypeVar;
val newTypeVars : int -> TypeVar list;
val newUnknown  : unit -> Type;
val isExplicit  : TypeVar -> bool;

val TypeOfTypeVar : TypeVar -> Type;
val fresh3DotType : unit -> RowType;
val contentsOfRowType : RowType -> Type Row * bool;
val isTupleType   : Type -> bool; 

val unify: Type -> Type -> unit;
val generalization: bool -> Location -> bool -> Type -> TypeScheme;
val specialization: TypeScheme -> Type;
val TypeOfScheme : TypeScheme -> Type;
val type_subst: (TypeVar * Type) list -> Type -> Type;
val freshSchemeOfType: Type -> TypeScheme;
val mkScheme: TypeVar list -> Type -> TypeScheme;
val trivial_scheme: Type -> TypeScheme;
val scheme_1u: (Type -> Type) -> TypeScheme;
val scheme_1u_eq: (Type -> Type) -> TypeScheme;
val scheme_1u_imp: (Type -> Type) -> TypeScheme;
val scheme_2u: (Type -> Type -> Type) -> TypeScheme;
val scheme_3u: (Type -> Type -> Type -> Type) -> TypeScheme;

val resetBindingLevel: unit -> unit;
val incrBindingLevel: unit -> unit;
val decrBindingLevel: unit -> unit;
val setCurrentBindingLevel: bool -> Type -> unit;

val makeEquality : Type -> unit
val typeViolatesEquality : Type -> bool;
val schemeViolatesEquality : TypeScheme -> bool;
val typeIsImperative : Type -> bool;

val sc_bool : TypeScheme;
val initial_bool_CE : ConEnv;
val initial_list_CE : ConEnv;
val initial_option_CE : ConEnv;
val initial_order_CE : ConEnv;
val initial_frag_CE : ConEnv;

val type_con : Type list -> TyName -> Type;
val type_arrow: Type -> Type -> Type;
val type_rigid_record : Type Row -> Type;
val type_flexible_record: Type Row -> RowType -> Type;
val type_ref: Type -> Type;
val type_pair: Type -> Type -> Type;
val type_product: Type list -> Type;

val type_unit      : Type;
val type_bool      : Type;
val type_int       : Type;
val type_word      : Type;
val type_word8     : Type;
val type_char      : Type;
val type_real      : Type;
val type_string    : Type;
val type_substring : Type;
val type_syserror  : Type;
val type_list      : Type -> Type;
val type_vector    : Type -> Type;
val type_exn       : Type;
val type_option    : Type -> Type;
val type_order     : Type;
val type_frag      : Type -> Type;
val type_ppstream  : Type;

val unit_General : CSig;

end;
