
open Misc List Fnlib Mixture Const Smlprim Globals Location Units;

datatype reason = 
    UnifyCircular | UnifyEquality | UnifyExplicit
  | UnifyTup | UnifyRec of Lab | UnifyOther

exception Unify of reason;

fun mktyname qual name info =
  {qualid={qual=qual, id=name}, info=ref info}
;
fun mkSML name info =
  mktyname "General" name info
;

(* Some predefined type names *)

val tyname_unit = mkSML "unit"      
    {tnStamp=0,  tnArity=0, tnEqu=TRUEequ,  tnStr=NILts}
and tyname_bool = mkSML "bool"      
    {tnStamp=0,  tnArity=0, tnEqu=TRUEequ,  tnStr=NILts}
and tyname_char = mktyname "Char" "char" 
    {tnStamp=0,  tnArity=0, tnEqu=TRUEequ,  tnStr=NILts}
and tyname_exn = mkSML "exn" 
    {tnStamp=0,  tnArity=0, tnEqu=FALSEequ, tnStr=NILts}
and tyname_frag = mkSML "frag"
    {tnStamp=0,  tnArity=1, tnEqu=TRUEequ,  tnStr=NILts}
and tyname_int = mktyname "Int" "int" 
    {tnStamp=0,  tnArity=0, tnEqu=TRUEequ,  tnStr=NILts}
and tyname_list = mkSML "list"  
    {tnStamp=0,  tnArity=1, tnEqu=TRUEequ,  tnStr=NILts}
and tyname_option = mkSML "option"
    {tnStamp=0,  tnArity=1, tnEqu=TRUEequ,  tnStr=NILts}
and tyname_order = mkSML "order"
    {tnStamp=0,  tnArity=0, tnEqu=TRUEequ,  tnStr=NILts}
and tyname_ppstream = mkSML "ppstream"  
    {tnStamp=0,  tnArity=0, tnEqu=FALSEequ, tnStr=NILts}
and tyname_real = mktyname "Real" "real" 
    {tnStamp=0,  tnArity=0, tnEqu=TRUEequ,  tnStr=NILts}
and tyname_ref = mkSML "ref"
    {tnStamp=0,  tnArity=1, tnEqu=REFequ,   tnStr=NILts}
and tyname_string = mktyname "String" "string" 
    {tnStamp=0,  tnArity=0, tnEqu=TRUEequ,  tnStr=NILts}
and tyname_substring = mktyname "Substring" "substring" 
    {tnStamp=0,  tnArity=0, tnEqu=FALSEequ,  tnStr=NILts}
and tyname_syserror = mktyname "OS" "syserror" 
    {tnStamp=0,  tnArity=0, tnEqu=FALSEequ,  tnStr=NILts}
and tyname_vector = mkSML "vector" 
    {tnStamp=0,  tnArity=1, tnEqu=TRUEequ,  tnStr=NILts}
and tyname_word = mktyname "Word" "word" 
    {tnStamp=0,  tnArity=0, tnEqu=TRUEequ,  tnStr=NILts}
and tyname_word8 = mktyname "Word8" "word" 
    {tnStamp=0,  tnArity=0, tnEqu=TRUEequ,  tnStr=NILts}
;

val tyname_bogus =
  mkSML "bogus" {tnStamp=(~1), tnArity=0, tnEqu=TRUEequ, tnStr=NILts};
val type_bogus = CONt([], tyname_bogus);
val sc_bogus = TypeScheme{ tscParameters=[], tscBody=type_bogus };

fun mkConInfo () =
  ref{ conArity=(~1), conSpan=(~1),
       conIsGreedy=false,
       conTag=(~1), conType=sc_bogus }
;

fun mkExConInfo () =
  ref{ exconArity=(~1), exconIsGreedy=false, exconTag=NONE }
;

fun mkPrimInfo arity prim =
  { primArity=arity, primOp=prim }
;

fun isExConStatic (ei : ExConInfo) =
  case #exconTag(!ei) of
      SOME _ => true
    | NONE => false
;

fun isNilRowType rho =
  case !rho of
      NILrow => true
    | VARrow _ => false
    | _ => fatalError "isNilRowType"
;

fun normalizeRecType (r: RecType) =
  case !(#rho(!r)) of
      NILrow => ()
    | VARrow _ => ()
    | LINKrow rho' =>
        (setRtRho r rho'; normalizeRecType r)
    | FIELDrow(lab, tau, rho') =>
        (setRtFields r (insertField (lab, tau) (#fields(!r)));
         setRtRho r rho'; normalizeRecType r)
;

fun normType tau =
  case tau of
      VARt var =>
        (case #tvKind(!var) of
             LinkTo tau' =>
               let val tau'' = normType tau' in
                 setTvKind var (LinkTo tau'');
                 tau''
               end
           |  _ => tau)
    | RECt r =>
        (normalizeRecType r; tau)
    | _ => tau;

fun isTupleType tau = 
    case normType tau of
	RECt rt => let val {fields, rho} = !rt
		    in isNilRowType rho andalso isTupleRow fields end
      | _       => false;

(* Printing types and type schemes *)

val variable_names = ref ([] : (TypeVar * string) list);
val variable_counter = ref 0;

fun collectExplicitVars tau =
  case normType tau of
    VARt var =>
      (case #tvKind(!var) of
           Explicit name =>
             if exists (fn(var',_) => var=var') (!variable_names) then ()
             else
		 (* Under value polymorphism, this code deletes underscores
                    in tyvars from signatures compiled with imperative types, 
                    such as those in the library. *)
		 let val newname = 
		     if !value_polymorphism then
			 case explode name of 
			     #"'" :: #"_" :: rest => 
			                         implode(#"'" :: rest)
			   | #"'" :: #"'" :: #"_" :: rest => 
						 implode(#"'" :: #"'" :: rest)
			   | _ => name
		     else
			 name
		 in
		     variable_names := (var, newname) :: !variable_names
		 end
         | _ => ())
  | ARROWt(t,t') =>
      (collectExplicitVars t; collectExplicitVars t')
  | CONt(ts, tn) =>
      app collectExplicitVars ts
  | RECt rt =>
      app_field collectExplicitVars (#fields(!rt))
;

fun alphaOfInt i =
  if i < 26 then
    CharVector.fromList [Char.chr (i + 97)]
  else
    alphaOfInt(i div 26) ^ alphaOfInt(i mod 26)
;

fun prTypeVar var =
  msgString (lookup var (!variable_names))
  handle Subscript =>
  let val {tvEqu, tvImp, ...} = !var
      val name =
        "'"^
        (if tvEqu then "'" else "") ^
        (if tvImp andalso not (!value_polymorphism) then "_" else "") ^
        alphaOfInt( !variable_counter )
  in
    incr variable_counter;
    if exists (fn (_,name') => name' = name) (!variable_names) then
      prTypeVar var
    else
      (variable_names := (var, name) :: !variable_names;
       msgString name)
  end;

fun arrowsToList (ARROWt(t, t')) = t :: arrowsToList t'
  | arrowsToList t = [t]
;

fun prType prior tau =
  let
    fun prParen prior' s = if prior >= prior' then msgString s else ()
  in
    case normType tau of
      VARt var =>
        prTypeVar var
    | ARROWt(t,t') =>
        let val ts = t :: arrowsToList t' in
          prParen 1 "("; msgIBlock 0;
          prTypeSeq 1 " ->" 0 ts;
          prParen 1 ")"; msgEBlock()
        end
    | CONt(ts, tn) =>
        (case ts of
             []  => ()
           | [t] => (prType 2 t; msgString " ")
           | _ =>
               (msgIBlock 0; msgString "(";
                prTypeSeq 0 "," 1 ts;
                msgString ") "; msgEBlock());
         printTQ tn)
    | RECt rt =>
        let val {fields=fs, rho=rho} = !rt in
          if isNilRowType rho then
            (if null fs then
               printTQ tyname_unit
             else if isTupleRow fs then
               (prParen 2 "(";
                msgIBlock 0;
                prTypeSeq 2 " *" 0 (map snd fs);
                prParen 2 ")"; msgEBlock())
             else
               (msgString "{"; msgIBlock 0;
                prTypeRow fs; msgString "}"; msgEBlock()))
          else if null fs then
            msgString "{...}"
          else
            (msgString "{"; msgIBlock 0;
             prTypeRow fs; msgString ","; msgBreak(1, 0);
             msgString "...}"; msgEBlock())
        end
  end

and prTypeSeq prior sep offset ts =
  case ts of
      [] => ()
    | [t] =>
        prType prior t
    | t :: rest =>
        (prType prior t; msgString sep; msgBreak(1, offset);
         prTypeSeq prior sep offset rest)

and prTypeRow fs =
  case fs of
      [] => ()
    | [(lab,t)] =>
        (printLab lab; msgString " :"; msgBreak(1, 2); prType 0 t)
    | (lab,t) :: rest =>
        (printLab lab; msgString " :"; msgBreak(1, 2); prType 0 t;
         msgString ","; msgBreak(1, 0); prTypeRow rest)
;

fun resetTypePrinter () =
(
  variable_names := [];
  variable_counter := 0
);

fun printType tau =
(
  resetTypePrinter();
  collectExplicitVars tau;
  prType 0 tau;
  resetTypePrinter()
);

fun printNextType tau =
  prType 0 tau
;

fun printScheme scheme =
  let val TypeScheme {tscBody, ...} = scheme
  in printType tscBody end
;

(* Error printing *)

fun errorToplevelImperativeVar loc var =
(
  msgIBlock 0;
  errLocation loc;
  if !value_polymorphism then
      errPrompt "Value polymorphism: Free type variable at top level"
  else
      errPrompt "Free imperative type variable at top level";
  (* printType var; *) msgEOL();
  msgEBlock();
  raise Toplevel
);

(* Binding levels *)

val binding_level = ref 0;

fun resetBindingLevel() = binding_level := 0;
fun incrBindingLevel() = incr binding_level;
fun decrBindingLevel() = decr binding_level;

fun setCurrentBindingLevel isOverloaded = fn
    VARt var =>
      (setTvOvl var isOverloaded;
       setTvLevel var (!binding_level))
  | _ => fatalError "setCurrentBindingLevel"
;

fun newExplicitTypeVar syntaxTypeVar =
  let val tv = ref
    {
      tvKind = Explicit syntaxTypeVar,
      tvLevel = !binding_level,
      tvImp = !value_polymorphism,
      tvEqu = false,
      tvOvl = false
    }
  in
    (case explode syntaxTypeVar of
        #"'" :: #"'" :: #"_" :: _ =>
          (setTvEqu tv true; setTvImp tv true)
      | #"'" :: #"'" :: _ =>
          setTvEqu tv true
      | #"'" :: #"_" :: _ =>
          setTvImp tv true
      | _ => ());
    tv
  end;

fun mkTypeVar equ imp ovl level = ref
{
  tvKind = NoLink,
  tvEqu = equ,
  tvImp = !value_polymorphism orelse imp,
  tvOvl = ovl,
  tvLevel = level
};

fun newTypeVar equ imp ovl =
  mkTypeVar equ imp ovl (!binding_level)
;

fun newTypeVars 0 = []
  | newTypeVars n = newTypeVar false false false :: newTypeVars (n-1)
;

fun refreshTypeVar (tv : TypeVar) =
  let val {tvEqu, tvImp, tvOvl, ...} = !tv
  in newTypeVar tvEqu tvImp tvOvl end
;

fun isExplicit (tv : TypeVar) =
  let val {tvKind, ...} = !tv in
    case tvKind of
        Explicit _ => true
      | _ => false
  end;

fun newUnknown() =
  VARt (newTypeVar false false false)
;

fun TypeOfTypeVar tv = VARt tv;

fun freshNilRowType() = ref NILrow;

fun mkVarRowType equ imp level =
  ref (VARrow (ref {rvEqu=equ, rvImp=imp, rvLevel=level}))
;

(* Correct binding level will be set later, *)
(* during type-checking *)

fun fresh3DotType() = mkVarRowType false false ~1;

fun contentsOfRowTypeAcc acc rho =
  case !rho of
      NILrow => (acc, false)
    | VARrow _ => (acc, true)
    | LINKrow rho' =>
        contentsOfRowTypeAcc acc rho'
    | FIELDrow(lab, tau, rho')  =>
        contentsOfRowTypeAcc ((lab,tau)::acc) rho'
;

fun contentsOfRowType rho =
  contentsOfRowTypeAcc [] rho
;

fun extractLab r r' (lab: Lab) (rho: RowType) =
  case !rho of
      NILrow => 
	  if isTupleType (RECt r) andalso isTupleType (RECt r') then
	      raise Unify UnifyTup
	  else 
	      raise Unify (UnifyRec lab)
    | VARrow rv =>
        let val {rvEqu=equ, rvImp=imp, rvLevel=level, ...} = !rv
            val rho' = mkVarRowType equ imp level
            val tau' = VARt (mkTypeVar equ imp false level)
        in
          rho := (FIELDrow(lab, tau', rho'));
          (tau', rho')
        end
    | LINKrow _ => fatalError "extractLab"
    | FIELDrow _ => fatalError "extractLab"
;

fun occur_check var tau =
  let fun test t =
    case normType t of
        VARt var' =>
          if var = var' then
            raise Unify UnifyCircular
          else ()
      | ARROWt(t1,t2) =>
          (test t1; test t2)
      | CONt(ts, _) =>
          app test ts
      | RECt (ref{fields, ...}) =>
          app_field test fields
  in test tau end;

fun pruneRho_level (max_level: int) (rho: RowType) =
  case !rho of
      NILrow => ()
    | VARrow rv =>
        if #rvLevel(!rv) > max_level then
          setRvLevel rv max_level
        else ()
    | LINKrow _ => fatalError "pruneRho_level"
    | FIELDrow _ => fatalError "pruneRho_level"
;

fun setRho_level (rho: RowType) =
  case !rho of
      VARrow rv =>
        setRvLevel rv (!binding_level)
    | _ => fatalError "setRho_level"
;

fun prune_level max_level =
  let fun prune t =
    case normType t of
        VARt var =>
          if #tvLevel(!var) > max_level then
            (if isExplicit var then
               raise Unify UnifyExplicit
             else
               setTvLevel var max_level)
          else ()
      | ARROWt(t1,t2) =>
          (prune t1; prune t2)
      | CONt(ts, _) =>
          app prune ts
      | RECt (ref {fields, rho}) =>
          (app_field prune fields;
           pruneRho_level max_level rho)
  in prune end
;

fun makeEqualityRho (rho: RowType) =
  case !rho of
      VARrow rv =>
        setRvEqu rv true
    | _ => ()
;

fun makeEquality t =
  case normType t of
    VARt var =>
      if #tvEqu(!var) then ()
      else if isExplicit var then
        raise Unify UnifyEquality
      else
        setTvEqu var true
  | ARROWt(t1,t2) =>
      raise Unify UnifyEquality
  | CONt(ts, tyname) =>
      (case #tnEqu(! (#info(tyname))) of
           FALSEequ => raise Unify UnifyEquality
         | TRUEequ =>
             app makeEquality ts
         | REFequ => ())
  | RECt (ref {fields, rho}) =>
      (app_field makeEquality fields;
       makeEqualityRho rho)
;

fun makeImperativeRho rho =
  case !rho of
      VARrow rv =>
        setRvImp rv true
    | _ => ()
;

fun makeImperative t =
  case normType t of
    VARt var =>
      if #tvImp(!var) then ()
      else if isExplicit var then
        raise Unify UnifyExplicit
      else
        setTvImp var true
  | ARROWt(t1,t2) =>
      (makeImperative t1; makeImperative t2)
  | CONt(ts, _) =>
      app makeImperative ts
  | RECt (ref {fields, rho}) =>
      (app_field makeImperative fields;
       makeImperativeRho rho)
;

fun makeOverloaded t =
  case normType t of
    VARt var =>
      if #tvOvl(!var) then ()
      else if isExplicit var then
        raise Unify UnifyExplicit
      else
        setTvOvl var true
  | ARROWt _ => ()
  | CONt _ => ()
  | RECt _ => ()
;

fun updateAttributes (var : TypeVar) t =
  let val {tvEqu, tvImp, tvOvl, ...} = !var in
    if tvEqu then makeEquality t else ();
    if tvImp andalso not (!value_polymorphism) then makeImperative t else ();
    if tvOvl then makeOverloaded t else ()
  end;

fun unifyRho rho1 rho2 =
  if rho1 = rho2 then () else
  case (!rho1, !rho2) of
      (NILrow, NILrow) => ()
    | (NILrow, VARrow _) =>
        rho2 := LINKrow rho1
    | (VARrow _, NILrow) =>
        rho1 := LINKrow rho2
    | (VARrow rv1, VARrow rv2) =>
        let val {rvEqu=rvEqu1, rvImp=rvImp1, rvLevel=rvLevel1, ...} = !rv1
            and {rvEqu=rvEqu2, rvImp=rvImp2, rvLevel=rvLevel2, ...} = !rv2
        in
          if rvLevel2 > rvLevel1 then
            setRvLevel rv2 rvLevel1
          else ();
          setRvEqu rv2 (rvEqu1 orelse rvEqu2);
          setRvImp rv2 (rvImp1 orelse rvImp2);
          rho1 := LINKrow rho2
        end
    | (_, _) => fatalError "unifyRow"
;

fun unify (tau1: Type) (tau2: Type) =
  let val tau1' = normType tau1
      and tau2' = normType tau2
  in
    case (tau1', tau2') of
        (VARt var1, VARt var2) =>
         (if var1 = var2 then () else
            (case (#tvKind(!var1), #tvKind(!var2)) of
                (Explicit _, Explicit _) =>
                  raise Unify UnifyExplicit
              | (Explicit _,_) =>
                  linkVarToType var2 tau1'
              | (_,_) =>
                  linkVarToType var1 tau2'))
      | (VARt var, _) =>
          if isExplicit var then raise Unify UnifyExplicit
          else linkVarToType var tau2'
      | (_, VARt var) =>
          if isExplicit var then raise Unify UnifyExplicit
          else linkVarToType var tau1'
      | (ARROWt(a,b), ARROWt(a',b')) =>
          (unify a a'; unify b b')
      | (CONt(ts, tn), CONt(ts', tn')) =>
          if (isEqTN tn tn') then unifySeq ts ts'
          else raise Unify UnifyOther
      | (RECt r, RECt r') =>
          if r = r' then () else
          (unifyRec r r' : unit)
      | (_, _) =>
          raise Unify UnifyOther
  end

and linkVarToType var tau =
(
  occur_check var tau;
  prune_level (#tvLevel(!var)) tau;
  updateAttributes var tau;
  setTvKind var (LinkTo tau)
)

and unifySeq (ts: Type list) (ts': Type list) = app2 unify ts ts'

and unifyRec (r : RecType) (r' : RecType) =
  let val {fields=fs,  rho=rho}  = !r
      val {fields=fs', rho=rho'} = !r'
      fun unifyRecAcc fs1 rho1 fs2 rho2 acc =
            ((case (fs1, fs2) of
                  ([], []) => (unifyRho rho1 rho2; (rev acc, rho1))
                | ((lab1,t1)::fs1', []) =>
                    let val (t2, rho2') = extractLab r r' lab1 rho2 in
                      unify t1 t2;
                      unifyRecAcc fs1' rho1 [] rho2' ((lab1,t1)::acc)
                    end
                | ([], (lab2,t2)::fs2') =>
                    let val (t1, rho1') = extractLab r r' lab2 rho1 in
                      unify t1 t2;
                      unifyRecAcc [] rho1' fs2' rho2 ((lab2,t2)::acc)
                    end
                | ((lab1,t1)::fs1', (lab2,t2)::fs2') =>
                   (if lt_lab lab1 lab2 then
                      let val (t2, rho2') = extractLab r r' lab1 rho2 in
                        unify t1 t2;
                        unifyRecAcc fs1' rho1 fs2 rho2' ((lab1,t1)::acc)
                      end
                    else if lt_lab lab2 lab1 then
                      let val (t1, rho1') = extractLab r r' lab2 rho1 in
                        unify t1 t2;
                        unifyRecAcc fs1 rho1' fs2' rho2 ((lab2,t2)::acc)
                      end
                    else (* lab1 = lab2 *)
                      (unify t1 t2;
                       unifyRecAcc fs1' rho1 fs2' rho2 ((lab1,t1)::acc))))
             : (Lab * Type) list * RowType)
      val (fs'', rho'') = unifyRecAcc fs rho fs' rho' []
  in
    setRtFields r fs''; setRtRho r rho'';
    setRtFields r' fs''; setRtRho r' rho''
  end;

fun generalization onTop loc isExpansive tau =
  let val parameters = ref []
      fun collect_parameters tau =
            case normType tau of
                VARt var =>
                  let val {tvImp, tvOvl, tvLevel, ...} = !var in
                    if member var (!parameters) then ()
                    else if tvLevel <= !binding_level then
                      (if onTop then
                         errorToplevelImperativeVar loc (VARt var)
                       else ())
                    else if tvOvl then
                      (if onTop then
			 (* Should not be generalized; will be resolved *)
                         () 
                       else
                         setTvLevel var (!binding_level))
                    else if tvImp andalso isExpansive then
                      (if onTop then
                         errorToplevelImperativeVar loc (VARt var)
                       else
                         setTvLevel var (!binding_level))
                    else
                      parameters := var :: !parameters
                  end
              | ARROWt(t,t') =>
                  (collect_parameters t; collect_parameters t')
              | CONt(ts, _) =>
                  app collect_parameters ts
              | RECt (ref{fields, ...}) =>
                  app_field collect_parameters fields
  in
    collect_parameters tau;
    TypeScheme {tscParameters = !parameters, tscBody = tau}
  end;

fun mkScheme pars tau =
  TypeScheme{ tscParameters=pars, tscBody=tau }
;

fun trivial_scheme tau = TypeScheme{tscParameters = [], tscBody = tau};

fun scheme_1u_attr equ imp builder =
  let val () = incrBindingLevel()
      val v = VARt (newTypeVar equ imp false)
      val tau = builder v
  in
    decrBindingLevel();
    generalization false nilLocation false tau
  end;

fun scheme_1u     builder = scheme_1u_attr false false builder;
fun scheme_1u_eq  builder = scheme_1u_attr true  false builder;
fun scheme_1u_imp builder = scheme_1u_attr false true  builder;

fun scheme_2u builder =
  let val () = incrBindingLevel()
      val a1 = newUnknown()
      val a2 = newUnknown()
      val tau = builder a1 a2
  in
    decrBindingLevel();
    generalization false nilLocation false tau
  end;

fun scheme_3u builder =
  let val () = incrBindingLevel()
      val a1 = newUnknown()
      val a2 = newUnknown()
      val a3 = newUnknown()
      val tau = builder a1 a2 a3
  in
    decrBindingLevel();
    generalization false nilLocation false tau
  end;

fun specialization scheme =
  let val TypeScheme{ tscParameters, tscBody } = scheme in
    case tscParameters of
      [] => tscBody
    | _ =>
        let val newUnknowns =
              map (fn var => (var, VARt(refreshTypeVar var))) tscParameters
            fun copy tau =
                  case normType tau of
                    t as (VARt var) =>
                      (lookup var newUnknowns
                       handle Subscript => t)
                  | ARROWt(t,t') =>
                      ARROWt(copy t, copy t')
                  | CONt(ts, tn) =>
                      CONt(map copy ts, tn)
                  | RECt (ref{fields, rho}) =>
                      RECt (ref{fields=map_fields copy fields, rho=rho})
        in copy tscBody end
  end;

fun TypeOfScheme (TypeScheme {tscBody, ...}) = tscBody;

fun type_subst UE tau =
  case normType tau of
      VARt var =>
        (lookup var UE
         handle Subscript => fatalError "type_subst: Unknown variable")
    | ARROWt(t,t') =>
        ARROWt(type_subst UE t, type_subst UE t')
    | CONt(ts, tn) =>
        CONt(map (type_subst UE) ts, tn)
    | RECt (ref{fields, rho}) =>
        RECt (ref{fields=map_fields (type_subst UE) fields, rho=rho})
;

fun freshSchemeOfType tau =
  let
    val parameters = ref []
    fun collect_parameters tau =
          case normType tau of
              VARt var =>
                if member var (!parameters) then () else
                parameters := var :: !parameters
            | ARROWt(t,t') =>
                (collect_parameters t; collect_parameters t')
            | CONt(ts, _) =>
                app collect_parameters ts
            | RECt (ref{fields, ...}) =>
                app_field collect_parameters fields
    val () = collect_parameters tau
    val vs = map (fn _ => newTypeVar false false false) (!parameters)
    val UE = Fnlib.map2 (fn var => fn var' => (var, VARt var'))
                        (!parameters) vs
  in
    TypeScheme{tscParameters = vs, tscBody = type_subst UE tau}
  end;

fun typeViolatesEquality tau =
  case normType tau of
      VARt _ => false
    | ARROWt _ => true
    | CONt(ts, tn) =>
        (case #tnEqu(! (#info tn)) of
             FALSEequ => true
           | TRUEequ  => exists typeViolatesEquality ts
           | REFequ   => false)
    | RECt (ref{fields, ...}) =>
        exists_field typeViolatesEquality fields
;

fun schemeViolatesEquality sch =
  let val TypeScheme {tscBody, ...} = sch in
    case normType tscBody of
        ARROWt(t,t') =>
          typeViolatesEquality t
      | _ => false
  end;

fun rhoIsImperative rho =
  case !rho of
      NILrow => true
    | VARrow (ref{rvImp, ...}) => rvImp
    | _ => fatalError "rhoIsImperative"
;

fun typeIsImperative tau =
  case normType tau of
      VARt (ref{tvImp, ...}) => tvImp
    | ARROWt(t,t') =>
        (typeIsImperative t andalso typeIsImperative t')
    | CONt(ts, _) =>
        all typeIsImperative ts
    | RECt (ref{fields, rho}) =>
        (all_fields typeIsImperative fields andalso rhoIsImperative rho)

;

(* Some predefined types *)

fun type_con ts tyname = CONt(ts, tyname)
and type_arrow t1 t2 = ARROWt(t1, t2)
and type_rigid_record fs =
  RECt (ref{fields=sortRow fs, rho=freshNilRowType()})
and type_flexible_record fs rho =
(
  setRho_level rho;
  RECt (ref{fields=sortRow fs, rho=rho})
);

fun type_pair t t' = type_rigid_record (mkPairRow t t');
fun type_product ts = type_rigid_record (mkTupleRow ts);

val type_bool      = type_con [] tyname_bool;
val type_char      = type_con [] tyname_char;
val type_exn       = type_con [] tyname_exn;
fun type_frag t    = type_con [t] tyname_frag;
val type_int       = type_con [] tyname_int;
fun type_list t    = type_con [t] tyname_list;
fun type_option t  = type_con [t] tyname_option;
val type_order     = type_con [] tyname_order;
val type_ppstream  = type_con [] tyname_ppstream;
val type_real      = type_con [] tyname_real;
fun type_ref t     = type_con [t] tyname_ref;
val type_string    = type_con [] tyname_string;
val type_substring = type_con [] tyname_substring;
val type_syserror  = type_con [] tyname_syserror;
val type_word      = type_con [] tyname_word;
val type_word8     = type_con [] tyname_word8;
val type_unit      = type_product [];
fun type_vector t  = type_con [t] tyname_vector;

val sc_bool = trivial_scheme type_bool;
val sc_order = trivial_scheme type_order;

val infoFalse = mkSML "false"
  { conArity=0,   conIsGreedy=false,   conTag=0,   conSpan=2,
    conType=sc_bool }
and infoTrue = mkSML "true"
  { conArity=0,   conIsGreedy=false,   conTag=1,   conSpan=2,
    conType=sc_bool }
and infoNil = mkSML "nil"
  { conArity=0,   conIsGreedy=false,   conTag=0,   conSpan=2,
    conType= scheme_1u (fn a => type_list a) }
and infoCons = mkSML "::"
  { conArity=2,   conIsGreedy=true,    conTag=1,   conSpan=2,
    conType= scheme_1u (fn a =>
      type_arrow (type_pair a (type_list a)) (type_list a)) }
and infoNONE = mkSML "NONE"
  { conArity=0,   conIsGreedy=false,   conTag=0,   conSpan=2,
    conType= scheme_1u (fn a => type_option a) }
and infoSOME = mkSML "SOME"
  { conArity=1,   conIsGreedy=false,   conTag=1,   conSpan=2,
    conType= scheme_1u (fn a =>
      type_arrow a (type_option a)) }
and infoLESS = mkSML "LESS"
  { conArity=0,   conIsGreedy=false,   conTag=0,   conSpan=3,
    conType=sc_order }
and infoEQUAL = mkSML "EQUAL"
  { conArity=0,   conIsGreedy=false,   conTag=1,   conSpan=3,
    conType=sc_order }
and infoGREATER = mkSML "GREATER"
  { conArity=0,   conIsGreedy=false,   conTag=2,   conSpan=3,
    conType=sc_order }
and infoQUOTE = mkSML "QUOTE"
  { conArity=1,   conIsGreedy=false,   conTag=0,   conSpan=2,
    conType= scheme_1u (fn a =>
      type_arrow type_string (type_frag a)) }
and infoANTIQUOTE = mkSML "ANTIQUOTE"
  { conArity=1,   conIsGreedy=false,   conTag=1,   conSpan=2,
    conType= scheme_1u (fn a =>
      type_arrow a (type_frag a)) }
;

val initial_bool_CE = [infoFalse, infoTrue];
val initial_list_CE = [infoNil, infoCons];
val initial_option_CE = [infoNONE, infoSOME];
val initial_order_CE = [infoLESS, infoEQUAL, infoGREATER];
val initial_frag_CE = [infoQUOTE, infoANTIQUOTE];

val unit_General = newSig "General";

val () = setConstructors unit_General 100 initial_bool_CE;
val () = setConstructors unit_General 200 initial_list_CE;
val () = setConstructors unit_General 300 initial_option_CE;
val () = setConstructors unit_General 500 initial_order_CE;
val () = setConstructors unit_General 600 initial_frag_CE;

val () = setTnStr (#info tyname_unit)      (TYPEts([], type_unit));
val () = setTnStr (#info tyname_bool)      (DATATYPEts 100);
val () = setTnStr (#info tyname_list)      (DATATYPEts 200);
val () = setTnStr (#info tyname_option)    (DATATYPEts 300);
val () = setTnStr (#info tyname_order)     (DATATYPEts 500);
val () = setTnStr (#info tyname_frag)      (DATATYPEts 600);
