local
  open Fnlib Mixture Const Globals Location;
in

type CSig =
{
  uName:       string,
  uConBasis:   (string, ConStatus) Hasht.t,
  uVarEnv:     (string, TypeScheme) Hasht.t,
  uTyEnv:      (string, TyName) Hasht.t,
  uDatatypes:  (int, ConEnv) Hasht.t,
  uStamp:      SigStamp option ref,
                    (* present if this signature comes from a .ui file *)
  uMentions:   (string, SigStamp) Hasht.t
};

val conBasisOfSig : CSig -> (string, ConStatus) Hasht.t;
val varEnvOfSig : CSig -> (string, TypeScheme) Hasht.t;
val tyEnvOfSig : CSig -> (string, TyName) Hasht.t;

type SigTable = (string, CSig) Hasht.t;

val pervSigTable : SigTable;
val currentSigTable : SigTable ref;
val newSig : string -> CSig;
val currentSig : CSig ref;
val currentRenEnv : (string, int) Hasht.t ref;
val hasSpecifiedSignature : bool ref;
val readSig : string -> CSig;
val findSig : Location -> string -> CSig;
val pervasiveInfixTable : (string, InfixStatus) Hasht.t;
val initPervasiveEnvironments : unit -> unit;
val findAndMentionSig : Location -> string -> CSig;
val initInitialEnvironments : unit -> unit;
val protectCurrentUnit : (unit -> 'a) -> unit;
val currentUnitName : unit -> string;
val mkGlobalName : string -> QualifiedIdent;
val mkGlobalInfo : string -> 'a -> 'a global;
val mkUniqueGlobalName : string * 'a -> QualifiedIdent * 'a;
val newTypeStamp : unit -> int;
val newExcStamp : unit -> int;
val newValStamp : unit -> int;
val findInfo : (CSig -> (string, 'a) Hasht.t) ->
       (string, 'a) Env -> Location -> QualifiedIdent -> 'a;

val findConstructors : CSig -> int -> ConEnv;
val setConstructors : CSig -> int -> ConEnv -> unit;
val registerConstructors : ConEnv -> int;

val isEqTN : TyName -> TyName -> bool;
val updateCurrentInfixBasis : (string, InfixStatus) Env -> unit;
val updateCurrentConBasis : (string, ConStatus) Env -> unit;
val extendCurrentConBasis : (string, ConStatus) Env -> unit;
val updateCurrentStaticVE : (string, TypeScheme) Env -> unit;
val updateCurrentStaticTE : (string, TyName) Env -> unit;
val extendCurrentStaticTE : (string, TyName) Env -> unit;
val mkGlobalInfixBasis : unit -> (string, InfixStatus) Env;
val mkGlobalConBasis : unit -> (string, ConStatus) Env;
val mkGlobalVE : unit -> (string, TypeScheme) Env;
val mkGlobalTE : unit -> (string, TyName) Env;
val execToplevelOpen : Location -> string -> unit;
val printVQ : QualifiedIdent -> unit;
val printTQ : TyName -> unit;
val startCompilingUnit : string -> unit;
val rectifySignature : unit ->
  (QualifiedIdent * (QualifiedIdent * int)) list * (string * int) list;

end;
